import bge
from math import sqrt
from .hex_math import *

SIZE = 2.0		#the length of a hex-side

WIDTH = SIZE * 2	#the width of a hex
HORIZ = WIDTH * 3/4	#horizontal distance between columns
HEIGHT = sqrt(3)/2 * WIDTH	#height of a hexagon, distance between hexes on a row

def Msg(sub, body):
	#a super-simple sendMessage function
	bge.logic.sendMessage(sub,body)
	
	
class Board:
	#an array of hex tiles
	def __init__(self,brush,incoming,mouse_over, mouse_click, width,height):
		self.scene = bge.logic.getCurrentScene()	#used for addObject calls
		self.brush = brush							#used as placement origin for addObject calls
		self.incoming = incoming					#general message reciever
		self.mouse_over = mouse_over				#mouse-over sensor
		self.mouse_click = mouse_click				#mouse click sensor
		
		self.cursor = self.scene.objects['cursor']	#cursor object (tracks mouse cursor on board)
		self.target = self.scene.objects['target']	#target object (tracks currently targeted/selected tile)
		self.cursor.removeParent()						#parenting no longer needed
		self.target.removeParent()						#
		
		self.width = width				#width/height of the board in tiles
		self.height = height
		
		self.map = {y: {x: Tile() for x in range(width)} for y in range(height)}	#a 'blank' map template
		
		self.cursorHex = None						#Hex coords for cursor position on board
		self.targetHex = None						#Hex coords for target position on board
		
		self.lineHexes = []							#list of cube coords for drawing lines between tiles
		self.lineHexPips = []						#pips associated with lineHexes
		
		self.areaHexes = []							#list of cube coords for drawing areas around a hex
		self.areaHexPips = []						#pips associated with areaHexes
		
		self.drawLine = False						#do we need to redraw the Line?
		self.drawArea = False						#do we need to redraw the Area?
		
		self.areaRange = 0			#placeholder for updating range
		
		
	#use this to access a tile on the map	
	def getMapTile(self,q,r):
		return self.map[q][r+q//2]					#refer to an off-set position on the map.
														
	def findLine(self,A,B):
		distance = int(A.getDistanceTo(B))
		Msg("DISTANCE", str(distance))
		self.lineHexes = cube_line(A.cube,B.cube)
		
		for pip in self.lineHexPips:
			pip.endObject()
		self.lineHexPips = []
		
		for cube in self.lineHexes:
			hex = cube.to_hex()
			self.brush.worldPosition = self.map[hex.q][hex.r].pip.worldPosition
			pip = self.scene.addObject('MARKER',self.brush)
			self.lineHexPips.append(pip)
			
	#finds an area 'r' from hex 'h' and draws pips on the board
	def findArea(self,h,r):
		for pip in self.areaHexPips:
			pip.endObject()					
		
		self.areaHexPips = []
		self.areaHexes = cube_area_draw(self.targetHex, r)
		
		for cube in self.areaHexes:
			hex = cube.to_hex()
			
			try:
				tile = self.getMapTile(hex.q,hex.r)
			except KeyError:
				continue
			
			self.brush.worldPosition = tile.pip.worldPosition
			pip = self.scene.addObject('MARKER', self.brush)
			self.areaHexPips.append(pip)
			
			
			
	#Board main-loop logic
	def main(self):
	
		if self.brush['range'] != self.areaRange:
			self.areaRange = self.brush['range']
			self.drawArea = True
			Msg("RANGE",str(self.areaRange))
	
		if self.mouse_over.positive:
			hit = self.mouse_over.hitObject
			tile = hit['tile']
			hex = Hex(tile.hex.q, tile.hex.r - tile.hex.q//2)
			if not self.cursorHex or hex.q != self.cursorHex.q or hex.r != self.cursorHex.r:
				self.cursorHex = hex
				self.drawLine = True
			if self.mouse_click.positive:
				self.targetHex = hex
				self.drawArea = True
		else:
			self.cursorHex = None
		
		if self.cursorHex:
			self.cursor.visible = True
			self.cursor.worldPosition = self.getMapTile(self.cursorHex.q,self.cursorHex.r).pip.worldPosition
		else:
			self.cursor.visible = False
		
		if self.targetHex:
			self.target.visible = True
			self.target.worldPosition = self.getMapTile(self.targetHex.q, self.targetHex.r).pip.worldPosition
			if self.drawArea:
				self.findArea(self.targetHex, self.areaRange)
				self.drawArea = False
		else:
			self.target.visible = False
			
		if self.cursorHex and self.targetHex:
			
			A = self.getMapTile(self.cursorHex.q,self.cursorHex.r)
			B = self.getMapTile(self.targetHex.q,self.targetHex.r)
			'''
			A = self.map[self.targetHex.q][self.targetHex.r]
			B = self.map[self.cursorHex.q][self.cursorHex.r]
			'''
			bge.render.drawLine(A.pip.worldPosition, B.pip.worldPosition, [1,0,0])
			self.findLine(A,B)
	#generate a map for the Board
	def buildGrid(self):
		self.brush.worldPosition = [0, self.height * HEIGHT, 0]
		
		for x in range(self.height):
			for y in range(self.width):
				m = getEven(x)
				
				self.brush.worldPosition.y = ( y*HEIGHT + (m*(HEIGHT/2)) )
				pip = self.scene.addObject('HEX', self.brush)
				
				tile = self.map[x][y]
				pip['tile'] = tile
				pip['tile'].pip = pip
				tile.hex.q = x
				tile.hex.r = y
				tile.cube = tile.hex.to_cube()
				
				pip.children['coord'].text = "{} | {}".format(tile.hex.q,tile.hex.r)
			self.brush.worldPosition.x = (x+1)*HORIZ
			
class Tile:
	#a game tile object
	def __init__(self,q=0,r=0):
		self.hex = Hex(q,r)
		self.cube = self.hex.to_cube()
		
	
	def getDistanceTo(self,h):
		#find distance between me and another tile
		return cube_distance(self.cube, h.cube)
			
class Hex:
	#a generic Hex object w/ axial coordinates
	def __init__(self,q,r):
		self.q = q
		self.r = r
	def to_cube(self):
		x = self.q
		z = self.r
		y = -x-z
		return Cube(x,y,z)
				
class Cube:
	#a generic Cube object w/ cubic coordinates
	def __init__(self,x,y,z):
		self.x = x
		self.y = y
		self.z = z
	def to_hex(self):
		q = self.x
		r = self.z
		return Hex(q,r)	

'''	
def cube_to_hex(h):
	q = h.x
	r = h.z
	return Hex(q,r)

def hex_to_cube(h):
	x = h.q
	z = h.r
	y = -x-z
	return Cube(x,y,z)
'''
	
def Go(cont):
	own = cont.owner
	if 'board' not in own:
		incoming = cont.sensors['incoming']
		mouse_over = cont.sensors['MouseOver']
		mouse_click = cont.sensors['MouseClick']
		board = own['board'] = Board(own, incoming, mouse_over, mouse_click, 6, 6)
		board.buildGrid()
	else:
		own['board'].main()
 
